/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <math.h>
#include <time.h>
#include <pthread.h>
using namespace std;

extern ostream& comm;//commentry;

#include "ErrorHandler.h"
#include "Macros.h"
#include "Miscellaneous.h"
#include "SignalHandler.h"
#include "CommandLine.h"
#include "Matrix.h"
#include "Sequence.h"
#include "Parameters.h"
#include "Likelihood.h"


#define SMALLDOUB 1e-320
#define ADDEXP 100
#define MULTEXP (exp(ADDEXP))

class precalc;
class mspace;

void dneg(void *d);

inline double log(long double x,void* onerror=NULL){
//wrapper for log function that checks its argument isn't too small,
//and deals with the problem if it is. Doesn't deal with case where x
//is too big.
  if(x>SMALLDOUB){
    return log((double) x);
  }else if(x<=0){
    char m[100];
    sprintf(m,"Attempting to take logarithm of non-positive value: %Lg",x);
    menuitem debneg("Debug.",dneg);
    menu neglog(2,stop,debneg);
    neglog.setpars(onerror);
    fatalError(m,0,&neglog);//Should add debugging menu.
    //In theory, It might be better to throw this, and let the calling
    //routine decide whether the error is fatal.
  }else{
    int i=0;
    for(;x<SMALLDOUB;i++){
      x*=MULTEXP;
    };
    return log((double) x)-i*ADDEXP;
  };
  return 0;//Can't reach this line.
};

class tree{
public:
  int n;
  int numed;
  long double *length;
  tree *subtree;
  char *label;
  long double *down;//basic downwards lists.
  long double *down2;//downwards lists from top of branch
  long double *down3;//transformed downwards lists
  long double *up;//basic upwards lists
  long double *up2;//transformed upwards lists
  long double *up3;//upwards lists from bottom of branch
  long double *extra;
  long double *extra2;
  long double *extra3;//Just for parameters
//public:
  tree(){n=0;length=NULL;subtree=NULL;label=NULL;down=NULL;down2=NULL;down3=NULL;up=NULL;up2=NULL;up3=NULL;extra=NULL;extra2=NULL;extra3=NULL;};
  tree(const tree& T){n=T.n;length=T.length;subtree=T.subtree;label=T.label;up=T.up;up2=T.up2;up3=T.up3;down=T.down;down2=T.down2;down3=T.down3;extra=T.extra;extra2=T.extra2;extra3=T.extra3;};
  void operator=(tree T){n=T.n;length=T.length;subtree=T.subtree;label=T.label;up=T.up;up2=T.up2;up3=T.up3;down=T.down;down2=T.down2;down3=T.down3;extra=T.extra;extra2=T.extra2;extra3=T.extra3;};
  tree(int l,const char *s);
  tree(const char *s);
  tree *clone(int n);
  void clone(tree *t);
  void remove();
  long double *putedgelengths(long double *x){long double *ret=x;for(int i=0;i<n;i++){*ret=*(length+i);ret=(subtree+i)->putedgelengths(ret+1);};return ret;};
  const long double *setedgelengths(const long double *x){const long double *ret=x;for(int i=0;i<n;i++){*(length+i)=*ret;ret=(subtree+i)->setedgelengths(ret+1);};return ret;};
  int edges() const {int l=n;for(int i=0;i<n;i++){l+=(subtree+i)->edges();};return l;};
  void filledges(){numed=n;for(int i=0;i<n;i++){(subtree+i)->filledges();numed+=(subtree+i)->numed;};};
  int leaves() const {if(n==0){return 1;}else{int l=0;for(int i=0;i<n;i++){l+=(subtree+i)->leaves();};return l;};};
  void print(ostream& out=cout) const;
  void print(long double *bl,ostream& out=cout) const;
  friend ostream &operator <<(ostream &o,const tree &T);
  int findlabel(char *lab) const;
  void liklistdown(const Factmatrix& f,const sequence *data,int i,const long double *ex,long double len=-1); 
                              //builds the list of likelihoods
  void liklistdownfast(const Factmatrix& f,const sequence *data,int i,const long double *ex,const Realmatrix& gis,long double len=-1); 
  void liklistup(const params &p,const Factmatrix& f,const long double *ex,long double *above=NULL,long double len=0);
                              //builds the lists of upwards likelihoods.
  void printlists(int level,int start,int u,int st=0,int ed=0) const;
  void printdata(ostream& out,sequence* d) const;
  void getlengths(long double*& out) const{for(int i=0;i<n;i++){*(out++)=*(length+i);(subtree+i)->getlengths(out);};};
  long double *getlengths() const{long double *ans=new long double[this->edges()];long double *a2=ans;this->getlengths(a2);return ans;};
  const long double** getdownlists() const;
  const long double** getuplists() const;
  void getdownlists(const long double** list) const;
  void getuplists(const long double** list) const;
  void getalldown3lists(long double* list) const;
  void getallup2lists(long double* list) const;
  void getdown3lists(long double* list) const;
  void getup2lists(long double* list) const;
  const tree *findedge(int e,long double &len) const;
  long double *hessian(const Factmatrix& f,const sequence *data,const params& p);
  long double *hessian_threaded(const Factmatrix& f,const sequence *data,const params& p,int st=-1);
  void sitehess(const Factmatrix& f,const sequence *data,const params& p,int i,long double *siteliks,const precalc &x,mspace &mem);
  long double *approxHessian(const Factmatrix& f,const sequence *data,const params& p);
  long double *derivs(const Factmatrix& f,const sequence *data,const params& p);
  long double likelihood(const Factmatrix& f,const sequence *data,const params& p);
  long double fastlikelihood(const Factmatrix& f,const sequence *data,const params& p);
  long double *sitelikelihoods(const Factmatrix& f,const sequence *data,const params& p);
  void shiftbranchlength(int no,long double eps){for(int i=0;i<n;i++){if(no--==0){*(length+i)+=eps;return;}else if(no<(subtree+i)->edges()){(subtree+i)->shiftbranchlength(no,eps);return;}else{no-=(subtree+i)->edges();};};};
  void printspecies(ostream& out,int* no) const;
  void liklistextrabl(const Factmatrix &f,int j,const params& p,int st,long double ul,const long double* ex);
  void liklistextrabl(const Factmatrix &f,const params& p,int st,long double ul,const long double *ex);
  void liklistextra(const Factmatrix& f,const params& p,const long double *ex,const long double *coeffs,const Realmatrix *dD,long double *T,long double *U);
              //builds the lists of extra likelihoods.
  void liklistextra(const Factmatrix& f,int i,const Realmatrix& M,const long double *ex,const long double *coeffs,long double *T,long double *U);
  void liklistextra(const Factmatrix& f,int i,const params& p,const long double *ex,long double len=0);
  long double secder(int i,int j,const params& p,const Factmatrix& f,long double ul,const Realmatrix& M,const Realmatrix& N,const long double *&ex,const long double *&coeffs)const;
  long double sdsame(int i,int j,const Factmatrix& f, const params &p,const Realmatrix& M,const Realmatrix& N,const Realmatrix& O,const long double *mat,const long double *mat2,const long double *mat3,const long double *mat4,const int *part,const long double* upvals,const long double* downvals,const long double *ex,const long double *l,const long double *FIJ,const long double *coeffs,const long double *coeffs2,const long double *coeffs3,const long double *coeffs4,const long double *coeffs5,const long double *lsq,const long double *T,const long double *U)const throw();
  long double secderblp(int i,int j,const params& p,const Factmatrix& f,const long double* ex,const Realmatrix& R,long double ul,const long double *upvals,const long double *coeffs,int offset=0);
  long double secderbl(int i,int j,const params& p,const Factmatrix& f,const long double *ex,long double ul=0,int offset=0);
  void secderbl(int i,const params& p,const Factmatrix& f,const long double *ex,long double *&output,int edg,long double ul=0,int offset=0);
  void setextra(int ed);
  //Routines for calculating directional derivatives.
  void liklistextrabldir(const Factmatrix &f,const params& p,int st,long double ul,const long double* ex,const long double *dir);
  void liklistextradir(const Factmatrix& f,const params& p,const long double *ex,const Realmatrix &M, const long double *dir);
              //builds the lists of extra likelihoods.
  void liklistextradir(const Factmatrix& f,const Realmatrix& M,const params& p,const long double *ex,long double len=0);
  long double secderdir(const params& p,const Factmatrix& f,long double ul,const Realmatrix& M,const long double *&ex,const long double *coeffs)const;
  long double sdsamedir(const Factmatrix& f, const params &p,const Realmatrix& M,const Realmatrix& O,const long double *mat,const long double *mat2,const long double *mat3,const long double *mat4,const int *part,const long double* upvals,const long double* downvals,const long double *ex,const long double *l,const long double *FIJ)const throw();
  long double secderblpdir(const long double *dir,const params& p,const Factmatrix& f,const long double* ex,const Realmatrix& R,long double ul,const long double *coeffs,int offset=0);
  long double secderbldir(const long double *dir,const params& p,const Factmatrix& f,const long double *ex,long double ul=0,int offset=0);
  long double *directionalDerivs(const Factmatrix& f,const sequence *data,const params& p,const long double *dir);
  void debugneg(long double val,ostream& out=cout,int siteno=-1,int l=0);
  void scale(long double x){for(int i=0;i<n;i++){(subtree+i)->scale(x);*(length+i)*=x;};};
  void setblzero();
  void parsimonybl(const Realmatrix &C,sequence *data);
  void parsimonyblsno(const Realmatrix &C,sequence *data,int m,long double inf);
  void parsimonyupdatebl(const Realmatrix& C,int i,long double );
  void loglengths();
  void makenonzerobranches(){for(int i=0;i<n;i++){if(ZEROST(*(length+i))){*(length+i)=1e-5;};(subtree+i)->makenonzerobranches();};};
  //For Mixture models
  long double *Mixhessian_threaded(const Factmatrix* f,const sequence *data,int num,const long double *prob,const params* p,int st=-1);
  long double Mixlikelihood(const Factmatrix* f,const sequence *data,int num,const long double *prob,const params* p);
};

void tree::loglengths(){
  for(int i=0;i<n;i++){
    if(*(length+i)<=1){
      *(length+i)=1e-6;
    }else{
      *(length+i)=log(*(length+i));
    };
    (subtree+i)->loglengths();
  };
};

tree *tree::clone(int m){//produces m copies of this tree, each with their own memory for lists.
  tree *ans=new tree[m];
  int ed=this->edges();
  for(int i=0;i<m;i++){
    this->clone(ans+i);
    (ans+i)->setextra(ed);
  };
  return ans;
};

void tree::clone(tree *t){
  t->n=n;
  t->numed=numed;
  if(n>0){
    t->subtree=new tree[n];
    t->length=new long double[n];
  };
  t->up=new long double[NumCodons];
  t->up2=new long double[NumCodons];
  t->up3=new long double[NumCodons];
  t->down=new long double[NumCodons];
  t->down2=new long double[NumCodons];
  t->down3=new long double[NumCodons];
  for(int i=0;i<n;i++){
    *(t->length+i)=*(length+i);
    (subtree+i)->clone(t->subtree+i);
  };
  if(label!=NULL){
    t->label=new char[strlen(label)+1];
    strcpy(t->label,label);
  };
};

void tree::getalldown3lists(long double* list) const{
  int pos=0;
  for(int i=0;i<n;i++){
    (subtree+i)->getdown3lists(list+pos*NumCodons);
    pos+=(subtree+i)->edges()+1;
  };
};

void tree::getallup2lists(long double* list) const{
  int pos=0;
  for(int i=0;i<n;i++){
    (subtree+i)->getup2lists(list+pos*NumCodons);
    pos+=(subtree+i)->edges()+1;
  };
};

void tree::getdown3lists(long double* list) const{
  for(int i=0;i<NumCodons;i++){
    *(list+i)=*(down3+i);
  };
  int pos=1;
  for(int i=0;i<n;i++){
    (subtree+i)->getdown3lists(list+pos*NumCodons);
    pos+=(subtree+i)->edges()+1;
  };
};

void tree::getup2lists(long double* list) const{//Can't run this on root node
  for(int i=0;i<NumCodons;i++){
    *(list+i)=*(up2+i);
  };
  int pos=1;
  for(int i=0;i<n;i++){
    (subtree+i)->getup2lists(list+pos*NumCodons);
    pos+=(subtree+i)->edges()+1;
  };
};

void tree::getdownlists(const long double** list) const{
  //Usually use this to fill lists created by the root node. Not valid
  //for root node.
  *list=down;
  int pos=1;
  for(int i=0;i<n;i++){
    (subtree+i)->getdownlists(list+pos);
    pos+=(subtree+i)->edges()+1;
  };
};

void tree::getuplists(const long double** list) const{
  //Usually use this to fill lists created by the root node. Not valid
  //for root node.
  *list=up;
  int pos=1;
  for(int i=0;i<n;i++){
    (subtree+i)->getuplists(list+pos);
    pos+=(subtree+i)->edges()+1;
  };
};

const long double** tree::getdownlists() const{
  //Fetch all the lists at once to allow multiple actions to be
  //vectorised, reducing cache misses.
  const long double **ans=new const long double*[this->edges()];
  int pos=0;
  for(int i=0;i<n;i++){
    (subtree+i)->getdownlists(ans+pos);
    pos+=(subtree+i)->edges()+1;
  };
  return ans;
};

const long double** tree::getuplists() const{
  //Fetch all the lists at once to allow multiple actions to be
  //vectorised, reducing cache misses.
  const long double **ans=new const long double*[this->edges()];
  int pos=0;
  for(int i=0;i<n;i++){
    (subtree+i)->getuplists(ans+pos);
    pos+=(subtree+i)->edges()+1;
  };
  return ans;
};

const tree *tree::findedge(int e,long double &len) const{
  if(e==0){
    len=*length;
    return subtree;
  };
  const tree *curr=this;
  for(int s=e;s>0;){
    for(int t=0;t<curr->n;t++){
      int ed=(curr->subtree+t)->numed;
      if(ed<s){
	s-=ed+1;
      }else{
	len=*(curr->length+t);
	curr=curr->subtree+t;
	t=-1;
	if(s==0){
	  break;
	};
	s--;
      };
    };
  };    //now curr points to the tree at the bottom of edge e. len is
	//the length of edge e.
  return curr;
};

void tree::remove(){
  for(int i=0;i<n;i++){
    (subtree+i)->remove();
  };
  if(n>0){
    delete[] subtree;
    delete[] length;
  };
  if(label!=NULL){
    delete[] label;
  };
  if(up!=NULL){
    delete[] up; 
    delete[] up2;
    delete[] up3;
    up=NULL;
  };
  delete[] down;
  delete[] down2;
  delete[] down3;
  delete[] extra;
  delete[] extra2;
  delete[] extra3;
};

#include "DNAroutines.cpp"

void tree::liklistdown(const Factmatrix &f,const sequence *data,int m,const long double *ex,long double len){
  int l=data->getLength();
  if(l<m){
    fatalError("Sequence too short.");
    //    cout<<"Sequence too short!\n\n";
    //    exit(0);//Should probably try to improve error handling.
  };
  int j=0;
  int pos=(len<=0)?0:1;
  for(int i=0;i<n;i++){
    long double ll=*(length+i);
    (subtree+i)->liklistdown(f,data+j,m,ex+pos*NumCodons,ll);
    j+=(subtree+i)->leaves();
    pos+=(subtree+i)->edges()+1;
  };
  for(int i=0;i<NumCodons;i++){
    *(down+i)=1;
    for(int j=0;j<n;j++){
      *(down+i)*=*((subtree+j)->down2+i);
    };
  };
  if(n==0){
    int c=codon((data->seq)+CodonLength*m);
    putcodon(c,down);
  };
  if(up!=NULL){
    f.gammainv.act(down,down3);
    Likelihoodact(f,down3,down2,len,ex);
  };
};

void tree::liklistdownfast(const Factmatrix &f,const sequence *data,int m,const long double *ex,const Realmatrix &gis,long double len){
  int l=data->getLength();
  if(l<m){
    fatalError("Sequence too short.");
    //    cout<<"Sequence too short!\n\n";
    //    exit(0);//Should probably try to improve error handling.
  };
  int j=0;
  int pos=(len<=0)?0:1;
  for(int i=0;i<n;i++){
    long double ll=*(length+i);
    (subtree+i)->liklistdownfast(f,data+j,m,ex+pos*NumCodons,gis,ll);
    j+=(subtree+i)->leaves();
    pos+=(subtree+i)->edges()+1;
  };
  for(int i=0;i<NumCodons;i++){
    *(down+i)=1;
  };
  long double *tmp=new long double[NumCodons];
  for(int j=0;j<n;j+=2){
    for(int i=0;i<NumCodons;i++){
      *(down3+i)=*((subtree+j)->down2+i);
    };
    if(j+1<n){
      for(int i=0;i<NumCodons;i++){
	*(down3+i)*=*((subtree+j+1)->down2+i);
      };
      gis.act(down3,tmp);
    }else{
      f.gamma.act(down3,tmp);
    };
    for(int i=0;i<NumCodons;i++){
      *(down3+i)*=*(tmp+i);
    };
  };
  if(n==0){
    int c=codon((data->seq)+CodonLength*m);
    putcodon(c,down);
  };
  if(len>0){
    f.gammainv.act(down,down3);
    for(int i=0;i<NumCodons;i++){
      *(down2+i)=*(down3+i)*(*(ex+i));
    };
  };
};

/*
New function should allow missing values.


void tree::liklistdown(const Factmatrix &f,const sequence *data,int m,const long double *ex,long double len){
  int l=data->getLength();
  if(l<m){
    cout<<"Sequence too short!\n\n";
    exit(0);//Should probably try to improve error handling.
  };
  int j=0;
  int pos=(len<=0)?0:1;
  for(int i=0;i<n;i++){
    long double ll=*(length+i);
    (subtree+i)->liklistdown(f,data+j,m,ex+pos*NumCodons,ll);
    j+=(subtree+i)->leaves();
    pos+=(subtree+i)->edges()+1;
  };
  for(int i=0;i<NumCodons;i++){
    *(down+i)=1;
    for(int j=0;j<n;j++){
      *(down+i)*=*((subtree+j)->down2+i);
    };
  };
  if(n==0){
    int c=codon((data->seq)+CodonLength*m);
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i==c)?1:0;
    };
  };
  if(len>0){
    f.gammainv.act(down,down3);
    Likelihoodact(f,down3,down2,len,ex);
  };
};
*/

void tree::liklistup(const params &p,const Factmatrix& f,const long double *ex,long double *above,long double len){
  //Call this with above=NULL - above is only for recursive calls.
  if(above==NULL){
    if(up!=NULL){
      delete[] up;
      delete[] up2;
      delete[] up3;
    };
    up=NULL;
    up2=NULL;
    up3=NULL;
  };
  //  long double *x=new long double[NumCodons];
  if(above!=NULL){
    for(int i=0;i<NumCodons;i++){
      *(up+i)=*(above+i);
      *(above+i)*=*(p.pi+i);
    };
    f.gamma.TranspAct(above,up2);
    Likelihoodact(f,up2,up3,len,ex);
  };
  long double *t=new long double[NumCodons];
  int pos=(above==NULL)?0:1;
  for(int i=0;i<n;i++){//for each subtree
    for(int k=0;k<NumCodons;k++){
      *(t+k)=1;
      for(int j=0;j<n;j++){
	if(j!=i){
	  *(t+k)*=*((subtree+j)->down2+k);
	};
      };
    };
    if(above!=NULL){//account for the tree above.
      for(int k=0;k<NumCodons;k++){
	*(t+k)*=*(up3+k);
      };
    };      
    (subtree+i)->liklistup(p,f,ex+pos*NumCodons,t,*(length+i));
    pos+=(subtree+i)->edges()+1;
  };    
  delete[] t;
  //    delete[] x;
};

int tree::findlabel(char *lab) const{
  if(n==0){
    if(strcmp(label,lab)==0){
      return 0;
    }else{
      return -1;
    };
  };
  int l=0;
  for(int i=0;i<n;i++){
    int s=(subtree+i)->findlabel(lab);
    if(s!=-1){
      return s+l;
    }else{
      l+=(subtree+i)->leaves();
    };
  };
  return -1;
};

void tree::setextra(int ed){
  for(int i=0;i<n;i++){
    (subtree+i)->setextra(ed);
  };
  extra=new long double[NumCodons*(ed+params::numpars)];
  extra2=new long double[NumCodons*(ed+params::numpars)];
  extra3=new long double[NumCodons*(params::numpars)];
};

void inline putcodonps(int c,long double *down,long double inf){
  if(c<61){
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i==c)?0:inf;
    };
  }else if(c<125){//second nucleotide unknown
    c-=61;
    c/=4;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i/4==c)?0:inf;
    };
  }else if(c<189){//first Nucleotide unknown
    c-=125;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=((i%4==c%4)&&(i/16==c/16))?0:inf;
    };
  }else if(c<253){//first two Nucleotides unknown
    c-=189;
    c/=16;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i/16==c)?0:inf;
    };
  }else if(c<317){//Third nucleotide unknown
    c-=253;      
    c%=16;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i%16==c)?0:inf;
    };
  }else if(c<381){//Third and second nucleotides unknown
    c-=317;
    c%=16;
    c/=4;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=((i%16)/4==c)?0:inf;
    };
  }else if(c<445){//Third and second nucleotides unknown
    c-=381;
    c%=4;
    for(int i=0;i<NumCodons;i++){
      *(down+i)=(i%4==c)?0:inf;
    };
  }else{//all unknown
    for(int i=0;i<NumCodons;i++){
      *(down+i)=0;
    };
  };
};

void tree::setblzero(){
  //Sets all branchlengths to zero - used at the start of the
  //parsimony algorithm.
  for(int i=0;i<n;i++){
    (subtree+i)->setblzero();
    *(length+i)=0;
  };
};

void tree::parsimonybl(const Realmatrix &C,sequence *data){
  //Uses parsimony, with cost matrix C to fit the branch lengths.
  int l=data->getLength();
  long double inf=0;
  for(int i=0;i<NumCodonsSq;i++){
    if(*(C.entries+i)>inf){
      inf=*(C.entries+i);
    };
  };
  inf+=1;
  inf*=(numed*numed+1);//probably not necessary, but best to be safe.
  this->setblzero();
  for(int m=0;m<l;m++){//For each site
    this->parsimonyblsno(C,data,m,inf);
    int small=0;
    long double sval=inf;
    for(int i=0;i<NumCodons;i++){
      if(*(down+i)<sval){
	sval=*(down+i);
	small=i;
      };
    };
    this->parsimonyupdatebl(C,small,inf);
  };
  this->scale(1/(long double)l);
  this->makenonzerobranches();
};

void tree::parsimonyblsno(const Realmatrix &C,sequence *data,int m,long double inf){
  int j=0;
  for(int i=0;i<n;i++){
    (subtree+i)->parsimonyblsno(C,data+j,m,inf);
    j+=(subtree+i)->leaves();
  };
  for(int i=0;i<NumCodons;i++){
    *(down+i)=0;
    for(int j=0;j<n;j++){
      long double ij=inf;
      for(int k=0;k<NumCodons;k++){
	long double pij=*((subtree+j)->down+k)+*(C.entries+i*NumCodons+k);
	if(pij<ij){
	  ij=pij;
	};
      };
      *(down+i)+=ij;
    };
  };
  if(n==0){
    int c=codon((data->seq)+CodonLength*m);
    putcodonps(c,down,inf);
  }; 
};

void tree::parsimonyupdatebl(const Realmatrix& C,int i,long double inf){
  for(int j=0;j<n;j++){
    long double ij=inf;
    int l=0;
    for(int k=0;k<NumCodons;k++){
      long double pij=*((subtree+j)->down+k)+*(C.entries+i*NumCodons+k);
      if(pij<ij){
	ij=pij;
	l=k;
      };
    };
    *(length+j)+=*(C.entries+i*NumCodons+l);
    (subtree+j)->parsimonyupdatebl(C,l,inf);
  };
};

#include "Hessian.cpp"

long double *tree::hessian(const Factmatrix& f,const sequence *data,const params& p){
  int *shst=(int *)vars.seekval("showeverysite");
  int showeverysite=(shst==NULL)?0:*shst;
  int e=this->edges();
  int l=e+params::numpars+1;
  l=l*(l+1)/2;// add space for the hessian
  int *part=partition(NumCodons,f.D.entries);
  const Realmatrix *dD=p.Derivmats();
  int i=0;
  long double **mats=new long double*[(params::numpars)*(params::numpars+1)*2];
  for(int j=0;j<params::numpars;j++){
    for(int k=j;k<params::numpars;k++){
      *(mats+4*i)=new long double[NumCodonsSq];    
      *(mats+4*i+1)=new long double[NumCodonsSq];    
      *(mats+4*i+2)=new long double[NumCodonsSq];    
      *(mats+4*i+3)=new long double[NumCodonsSq];    
      makematrices(*(dD+j),*(dD+k),f.D.entries,part,*(mats+4*i),*(mats+4*i+1),*(mats+4*i+2),*(mats+4*i+3));
      i++;
    };
  };
  long double *T=new long double[e*params::numpars*NumCodons];
  long double *U=new long double[e*params::numpars*NumCodons];
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp(*(le+i)*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  long double *FIJ=new long double[NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int j=0;j<NumCodons;j++){
      *(FIJ+i*NumCodons+j)=(*(part+i)==*(part+j))?0:1/(*(f.D.entries+i)-*(f.D.entries+j));
    };
  };
  long double *coeffs=new long double[e*NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int k=0;k<NumCodons;k++){
      if(*(part+i)==*(part+k)){
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(le+ed)*(*(ex+ed*NumCodons+i));
	};
      }else{
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(FIJ+i*NumCodons+k)*(*(ex+ed*NumCodons+i)-(*(ex+ed*NumCodons+k)));
	};
      };
    };
  };
  long double *coeffsums=new long double[NumCodonsSq];
  long double *coeffsums2=new long double[NumCodonsSq];
  long double *coeffsums3=new long double[NumCodonsSq];
  long double *coeffsums4=new long double[NumCodonsSq];
  long double *coeffsums5=new long double[NumCodonsSq];
  long double *lsq=new long double[NumCodons];
  int sites=data->length;
  long double *ans=new long double[l];
  for(int i=0;i<l;i++){
    *(ans+i)=0;
  };
  long double *siteliks=new long double[l];
//Store values currently being calculated here, to avoid corrupting
//the state in case of interruption
  i=0;
  if(Recover!=NULL&&!strcmp(Recover->retid,"hessian")){//Trying to recover
    comm<<"Recovering from saved file.\n";
    Recover->load(2,1,&i,l,ans);
    comm<<"Starting from site "<<i<<".\n";
  };
  int stv1=CurrentState->addvar('i',1,&i);
  CurrentState->addvar('L',l,ans,"hessian");
  for(;i<sites;i++){//for each site.
    if(!((i+1)%showeverysite)){//should make this depend on the number of sites
      comm<<"Working on site "<<i+1<<" of "<<sites<<":\n";
    };
    this->liklistdown(f,data,i,ex);
    this->liklistup(p,f,ex,NULL);
    //Start with likelihood
    *(siteliks)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks)+=*(p.pi+j)*(*(down+j));
    };
    for(int j=1+e;j<l;j++){
      *(siteliks+j)=0;
    };
    this->liklistextra(f,p,ex,coeffs,dD,T,U);
    //Actually, the following few lines should be able to be run once for all parameters, saving time.
    long double *upvals=new long double[e*NumCodons]; 
    long double *downvals=new long double[e*NumCodons]; 
    /*    const long double **tmp=this->getuplists();
    f.gamma.MultiTranspActProd(e,tmp,p.pi,upvals);//Actually, these can be replaced.
    delete[] tmp;
    tmp=this->getdownlists();
    f.gammainv.MultiAct(e,tmp,downvals);
    delete[] tmp;  */
    this->getallup2lists(upvals);
    this->getalldown3lists(downvals);
    for(int j=0;j<NumCodons;j++){
      for(int k=0;k<NumCodons;k++){
	*(coeffsums+j*NumCodons+k)=0;
	*(coeffsums2+j*NumCodons+k)=0;
	*(coeffsums3+j*NumCodons+k)=0;
	*(coeffsums4+j*NumCodons+k)=0;
	*(coeffsums5+j*NumCodons+k)=0;
      };
      *(lsq+j)=0;
    };
    for(int ed=0;ed<e;ed++){	
      for(int j=0;j<NumCodons;j++){
	for(int k=0;k<NumCodons;k++){//These loops take up about 5% total time
	  long double w=(*(upvals+ed*NumCodons+j))*(*(downvals+ed*NumCodons+k));
	  *(coeffsums+j*NumCodons+k)+=*(coeffs+ed*NumCodonsSq+j*NumCodons+k)*w;
	  long double t=(*(ex+ed*NumCodons+j))*w;
	  *(coeffsums2+j*NumCodons+k)+=t;
	  *(coeffsums3+j*NumCodons+k)+=t*(*(le+ed));
	  long double u=(*(ex+ed*NumCodons+k))*w;
	  *(coeffsums4+j*NumCodons+k)+=u;
	  *(coeffsums5+j*NumCodons+k)+=u*(*(le+ed));
	  if(*(part+j)==*(part+k)){
	    *(lsq+j)+=*(le+ed)*(*(le+ed))*(*(ex+ed*NumCodons+j))*w;
	  };
	};
      };
    };
  //Up to here, this can all be precalculated.
    long double *jj=siteliks+1+e+params::numpars;
    for(int j=0;j<e;j++){
      //Second derivative with respect to branch j
      long double len=*(le+j);
      const tree *curr=this->findedge(j,len);
      *(jj++)=d2liketact(f,curr->up2,curr->down3,len,p,ex+j*NumCodons);

      this->secderbl(j,p,f,ex,jj,e,0);
      for(int k=0;k<params::numpars;k++){
	//Derivative with respect to branch j and parameter k
	*(jj++)=this->secderblp(j,k,p,f,ex,*(dD+k),0,upvals,coeffs);
      };
    };
    long double *acc=siteliks+1+e+params::numpars+e*params::numpars+e*(e+1)/2;
    int ii=0;
    for(int j=0;j<params::numpars;j++){
      for(int k=j;k<params::numpars;k++){
	const long double *ext=ex;
	const long double *cfs=coeffs;
	*(acc+ii)=this->secder(j,k,p,f,0,*(dD+j),*(dD+k),ext,cfs);
	*(acc+ii)+=this->sdsame(j,k,f,p,*(dD+j),*(dD+k),p.d2D(j,k),*(mats+4*ii),*(mats+4*ii+1),*(mats+4*ii+2),*(mats+4*ii+3),part,upvals,downvals,ex,le,FIJ,coeffs,coeffsums2,coeffsums3,coeffsums4,coeffsums5,lsq,T,U);
	long double *ent=p.d2D(j,k).entries;
	for(int jj=0;jj<NumCodonsSq;jj++){
	  *(acc+ii)+=*(ent+jj)*(*(coeffsums+jj));
	};
	ii++;
      };
    };//Need to take derivatives with respect to branch lengths.
    //    comm<<"Derivatives with respect to branchlengths are:\n";
    for(int j=0;j<e;j++){//for each edge
      long double len;
      this->findedge(j,len);
      *(siteliks+j+1)=dliketact2(f,upvals+j*NumCodons,downvals+j*NumCodons,len,p,ex+j*NumCodons);
      //      comm<<*(siteliks+j+1)<<"\t";
      //      *(siteliks+j+1)=dliketact(f,curr->down3,curr->up2,len,p,ex+j*NumCodons);
    };
    delete[] upvals;
    delete[] downvals;
    for(int k=0;k<params::numpars;k++){
      //Calculate first derivatives with respect to parameters.
      long double *lks=new long double[NumCodons];
      for(int kk=0;kk<NumCodons;kk++){
	*(lks+kk)=0;
      };
      for(int kk=0;kk<NumCodons;kk++){
	for(int ii=0;ii<n;ii++){
	  long double th=1;	  
	  for(int jj=0;jj<n;jj++){
	    if(jj==ii){	      
	      th*=*((subtree+ii)->extra+k*NumCodons+kk);
	    }else{
	      th*=*((subtree+jj)->down2+kk);
	    };
	  };
	  *(lks+kk)+=th;
	};
      };
      for(int kk=0;kk<NumCodons;kk++){
	*(siteliks+1+e+k)+=*(lks+kk)*(*(p.pi+kk));
      };
      delete[] lks;
    };
    //Block Signals
    sigset_t all;
    sigset_t old;
    sigfillset(&all);
    sigprocmask(SIG_BLOCK,&all,&old);//Block signals while doing this
    //Now update answer
    try{
      *ans+=log(*(siteliks));
    }catch(int neg){
      //Shouldn't happen now.
      cerr<<"site likelihood "<<*(siteliks)<<"calculated. Aborting.\n";
      CurrentState->save();
      exit(1);
    };
    for(int i=1;i<1+e+params::numpars;i++){
      *(ans+i)+=(*(siteliks+i))/(*(siteliks));
    };
    int v=1+e+params::numpars;
    for(int j=0;j<e+params::numpars;j++){
      for(int k=j;k<e+params::numpars;k++){
	*(ans+v)+=(*(siteliks+v))/(*(siteliks));
	*(ans+v)-=(*(siteliks+j+1))*(*(siteliks+k+1))/(*(siteliks)*(*(siteliks)));
	v++;
      };
    };
    //Restore Signals
    sigprocmask(SIG_SETMASK,&old,NULL);//Unblock signals again
  };
  delete[] ex;
  delete[] le;
  //Now combine sitewise results into overall likelihoods.
  for(i=0;i<(params::numpars)*(params::numpars+1)*2;i++){
    delete[] *(mats+i);
  };
  delete[] coeffs;
  delete[] coeffsums;
  delete[] coeffsums2;
  delete[] coeffsums3;
  delete[] coeffsums4;
  delete[] coeffsums5;
  delete[] lsq;
  delete[] mats;
  delete[] part;
  delete[] FIJ;
  delete[] T;
  delete[] U;
  //  CurrentState->addvar('L',l,ans);
  CurrentState->remvar(stv1,"endhessian",2);
  delete[] siteliks;
  return ans;
};

int currentpos=0;
pthread_mutex_t posmut=PTHREAD_MUTEX_INITIALIZER;
int numthreads=0;
//int stop=100;

void setnumthreads(int i){
  numthreads=i;
};

void setpos(int i){
  currentpos=i;
};

inline int getnext(){
  pthread_mutex_lock(&posmut);
  int i=currentpos++;
  pthread_mutex_unlock(&posmut);
  return i;
};

class alldata{
public:
  tree *t;
  const Factmatrix *f;
  const sequence *data;
  const params *p;
  alldata(){t=NULL;f=NULL;data=NULL;p=NULL;};
  alldata(tree &tr,const Factmatrix & fa,const sequence *d,const params &pa){t=&tr;f=&fa;data=d;p=&pa;};
  void assign(tree &tr,const Factmatrix & fa,const sequence *d,const params &pa){t=&tr;f=&fa;data=d;p=&pa;};
};

void *callhessthreaded(void *d){
  alldata *a=(alldata *)d;
  return (a->t)->hessian_threaded(*(a->f),a->data,*(a->p));
};

class precalc{//stores all the precalculated data.
public:
  int e;
  int l;
  long double **mats;
  const Realmatrix *dD;
  long double *ex;
  long double *le;
  long double *FIJ;
  int *part;
  long double *coeffs;
  precalc(){e=0;l=0;mats=NULL;dD=NULL;ex=NULL;le=NULL;FIJ=NULL;part=NULL;coeffs=NULL;};
  precalc(const tree &t,const Factmatrix& f,const sequence *data,const params& p);
  void calc(const tree &t,const Factmatrix& f,const sequence *data,const params& p);
  ~precalc();
};


precalc::~precalc(){
  delete[] ex;
  delete[] le;
  for(int i=0;i<(params::numpars)*(params::numpars+1)*2;i++){
    delete[] *(mats+i);
  };
  delete[] mats;
  delete[] part;
  delete[] FIJ;
  delete[] coeffs;
};

precalc::precalc(const tree &t,const Factmatrix& f,const sequence *data,const params& p){
  //Should really make a threaded version of this.
  e=t.edges();
  l=e+params::numpars+1;
  l=l*(l+1)/2;// add space for the hessian
  part=partition(NumCodons,f.D.entries);
  dD=p.Derivmats();
  mats=new long double*[(params::numpars)*(params::numpars+1)*2];
  int i=0;
  for(int j=0;j<params::numpars;j++){
    for(int k=j;k<params::numpars;k++){
      *(mats+4*i)=new long double[NumCodonsSq];    
      *(mats+4*i+1)=new long double[NumCodonsSq];    
      *(mats+4*i+2)=new long double[NumCodonsSq];    
      *(mats+4*i+3)=new long double[NumCodonsSq];    
      makematrices(*(dD+j),*(dD+k),f.D.entries,part,*(mats+4*i),*(mats+4*i+1),*(mats+4*i+2),*(mats+4*i+3));
      i++;
    };
  };
  ex=new long double[NumCodons*e];
  le=t.getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  FIJ=new long double[NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int j=0;j<NumCodons;j++){
      *(FIJ+i*NumCodons+j)=(*(part+i)==*(part+j))?0:1/(*(f.D.entries+i)-*(f.D.entries+j));
    };
  };
  coeffs=new long double[e*NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int k=0;k<NumCodons;k++){
      if(*(part+i)==*(part+k)){
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(le+ed)*(*(ex+ed*NumCodons+i));
	};
      }else{
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(FIJ+i*NumCodons+k)*(*(ex+ed*NumCodons+i)-(*(ex+ed*NumCodons+k)));
	};
      };
    };
  };
};

void precalc::calc(const tree &t,const Factmatrix& f,const sequence *data,const params& p){
  //Should really make a threaded version of this.
  e=t.edges();
  l=e+params::numpars+1;
  l=l*(l+1)/2;// add space for the hessian
  part=partition(NumCodons,f.D.entries);
  dD=p.Derivmats();
  mats=new long double*[(params::numpars)*(params::numpars+1)*2];
  int i=0;
  for(int j=0;j<params::numpars;j++){
    for(int k=j;k<params::numpars;k++){
      *(mats+4*i)=new long double[NumCodonsSq];    
      *(mats+4*i+1)=new long double[NumCodonsSq];    
      *(mats+4*i+2)=new long double[NumCodonsSq];    
      *(mats+4*i+3)=new long double[NumCodonsSq];    
      makematrices(*(dD+j),*(dD+k),f.D.entries,part,*(mats+4*i),*(mats+4*i+1),*(mats+4*i+2),*(mats+4*i+3));
      i++;
    };
  };
  ex=new long double[NumCodons*e];
  le=t.getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  FIJ=new long double[NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int j=0;j<NumCodons;j++){
      *(FIJ+i*NumCodons+j)=(*(part+i)==*(part+j))?0:1/(*(f.D.entries+i)-*(f.D.entries+j));
    };
  };
  coeffs=new long double[e*NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int k=0;k<NumCodons;k++){
      if(*(part+i)==*(part+k)){
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(le+ed)*(*(ex+ed*NumCodons+i));
	};
      }else{
	for(int ed=0;ed<e;ed++){
	  *(coeffs+ed*NumCodonsSq+i*NumCodons+k)=*(FIJ+i*NumCodons+k)*(*(ex+ed*NumCodons+i)-(*(ex+ed*NumCodons+k)));
	};
      };
    };
  };
};

class mspace{
public:
  long double *coeffsums;
  long double *coeffsums2;
  long double *coeffsums3;
  long double *coeffsums4;
  long double *coeffsums5;
  long double *lsq;
  long double *T;
  long double *U;
  mspace(int e){coeffsums=new long double[NumCodonsSq];coeffsums2=new long double[NumCodonsSq];coeffsums3=new long double[NumCodonsSq];coeffsums4=new long double[NumCodonsSq];coeffsums5=new long double[NumCodonsSq];lsq=new long double[NumCodons];T=new long double[e*params::numpars*NumCodons];U=new long double[e*params::numpars*NumCodons];};
  ~mspace(){delete[] coeffsums;delete[] coeffsums2;delete[] coeffsums3;delete[] coeffsums4;delete[] coeffsums5;delete[] lsq;delete[] T;delete[] U;};
};

long double *recovercalchess(tree &t, const Factmatrix& f,const sequence *data,const params& p){
  //resumes an unfinished hessian calculation.
  //should already have loaded x. Now just need to load partial calculation.
  int e=t.edges();
  int l=e+params::numpars+1;
  l=l*(l+1)/2;// add space for the hessian
  int nth=Recover->num('i');//Number of threads that were running in interrupted program.
  int *torepeat=new int[nth];
  long double *answer=new long double[l];
  for(int i=0;i<l;i++){
    *(answer+i)=0;
  };
  long double *temp=new long double[l];
  for(int i=0;i<nth;i++){
    Recover->load(2,1,torepeat+i,l,temp);
    for(int j=0;j<l;j++){
      *(answer+j)+=*(temp+j);
    };
  };
  currentpos=0;
  precalc x(t,f,data,p);
  mspace mem(x.e);
  for(int i=0;i<nth;i++){
    t.sitehess(f,data,p,*(torepeat+i),temp,x,mem);   
    for(int j=0;j<l;j++){
      *(answer+j)+=*(temp+j);
    };
    if(*(torepeat+i)>currentpos){
      currentpos=*(torepeat+i);
    };
  };
  currentpos++;
  //Finished off the sites that were only partially calculated.
  if(numthreads==0){
    numthreads=NUMTHREADS;
  };
  char msg[40];
  sprintf(msg,"Producing %i threads for computation.",numthreads);
  info(msg,msgcode(2,0));
  //  comm<<"Producing "<<numthreads<<" threads for computation.\n";
  //  if(numthreads==1){ 
  //      return t.hessian(f,data,p); 
  //  }; 

  //  Can't use normal 1-threaded hessian, as it doesn't allow an
  //  arbitrary start point.
  pthread_t *worker=new pthread_t[numthreads-1];
  tree *tcopy=t.clone(numthreads-1);
  int i=0;
  int ecode=0;
  alldata *a=new alldata[numthreads-1];
  for(;i<numthreads-1;i++){
    (a+i)->assign(*(tcopy+i),f,data,p);
  };
  for(i=0;i<numthreads-1&&!ecode;i++){
    //for(i=0;i<1&&!ecode;i++){
    ecode=pthread_create((worker+i),NULL,callhessthreaded,a+i);
  };//Now i stores the actual number of pthreads created.
  if(ecode){
    sprintf(msg,"only able to produce %i threads.",i);
    warning(msg);
  };
  long double *ans=t.hessian_threaded(f,data,p);
  for(int j=0;j<l;j++){
    *(answer+j)+=*(ans+j);
  };
  ecode=0;
  for(int j=0;j<i;j++){
    //    comm<<"Trying to rejoin thread "<<j+1<<".\n";
    ecode=pthread_join(*(worker+j),(void**)&temp);
    if(ecode){
      char x[60];
      sprintf(x,"Threading error rejoining thread number %d",j);
      fatalError(x);
      //      cerr<<"Threading error rejoining thread number "<<j<<".\n";
      //      exit(1);
    };
    for(int k=0;k<l;k++){
      *(answer+k)+=*(temp+k);
    };
  };
  delete[] temp;
  delete[] ans;
  /*
  stop=-1;
  for(i=1;i<2&&!ecode;i++){
    ecode=pthread_create((worker+i),NULL,callhessthreaded,a+i);
  };//Now i stores the actual number of pthreads created.
  //  long double *ans=t.hessian_threaded(f,data,p);
  //  long double *temp;
  ecode=0;
  for(int j=1;j<i;j++){
    comm<<"Trying to rejoin thread "<<j+1<<".\n";
    ecode=pthread_join(*(worker+j),(void**)&temp);
    if(ecode){
      cerr<<"Threading error rejoining thread number "<<j<<".\n";
      exit(1);
    };
    for(int k=0;k<l;k++){
      comm<<"adding value "<<k<<".\n";
      *(ans+k)+=*(temp+k);
    };
    delete[] temp;
  };
  */

  for(i=0;i<numthreads-1;i++){
    (tcopy+i)->remove();
  };
  delete[] tcopy;
  delete[] worker;
  delete[] a;
  return answer;
};

long double *calchess(tree &t, const Factmatrix& f,const sequence *data,const params& p){
  int e=t.edges();
  int l=e+params::numpars+1;
  l=l*(l+1)/2;// add space for the hessian
  currentpos=0;
  if(numthreads==0){
    numthreads=NUMTHREADS;
  };
  char msg[40];
  sprintf(msg,"Producing %i threads for computation.",numthreads);
  info(msg,msgcode(2,0));
  //  comm<<"Producing "<<numthreads<<" threads for computation.\n";
  if(numthreads==1){
    return t.hessian(f,data,p);
  };
  pthread_t *worker=new pthread_t[numthreads-1];
  tree *tcopy=t.clone(numthreads-1);
  int i=0;
  int ecode=0;
  alldata *a=new alldata[numthreads-1];
  for(;i<numthreads-1;i++){
    (a+i)->assign(*(tcopy+i),f,data,p);
  };
  for(i=0;i<numthreads-1&&!ecode;i++){
    //for(i=0;i<1&&!ecode;i++){
    ecode=pthread_create((worker+i),NULL,callhessthreaded,a+i);
  };//Now i stores the actual number of pthreads created.
  if(ecode){
    sprintf(msg,"only able to produce %i threads.",i);
    warning(msg);
  };
  long double *ans=t.hessian_threaded(f,data,p);
  long double *temp;
  ecode=0;
  for(int j=0;j<i;j++){
    //    comm<<"Trying to rejoin thread "<<j+1<<".\n";
    ecode=pthread_join(*(worker+j),(void**)&temp);
    if(ecode){
      char x[60];
      sprintf(x,"Threading error rejoining thread number %d",j);
      fatalError(x);
      //      cerr<<"Threading error rejoining thread number "<<j<<".\n";
      //      exit(1);
    };
    for(int k=0;k<l;k++){
      *(ans+k)+=*(temp+k);
    };
    delete[] temp;
  };
  for(i=0;i<numthreads-1;i++){
    (tcopy+i)->remove();
  };
  delete[] tcopy;
  delete[] worker;
  delete[] a;
  return ans;
};

void accumulate(const long double *siteliks,long double *ans,int e,int l,void *dbn=NULL){
  //Used to add the calculated likelihoods for the current site to the
  //cumulative totals.
  try{
    *ans+=log(*(siteliks),dbn);
  }catch(int neg){
    //This shouldn't happen now.
    cerr<<"Thread "<<pthread_self()<<":\n";
    cerr<<"site likelihood "<<*(siteliks)<<"calculated. Aborting.\n";
    CurrentState->save();
    exit(1);
  };
  for(int j=1;j<1+e+params::numpars;j++){
    *(ans+j)+=(*(siteliks+j))/(*(siteliks));
  };
  int v=1+e+params::numpars;
  for(int j=0;j<e+params::numpars;j++){
    for(int k=j;k<e+params::numpars;k++){
      *(ans+v)+=(*(siteliks+v))/(*(siteliks));
      *(ans+v)-=(*(siteliks+j+1))*(*(siteliks+k+1))/(*(siteliks)*(*(siteliks)));
      v++;
    };
  };
};

long double *tree::hessian_threaded(const Factmatrix& f,const sequence *data,const params& p,int st){
  int *shst=(int *)vars.seekval("showeverysite");
  int showeverysite=(shst==NULL)?0:*shst;

  precalc x(*this,f,data,p);
  int i=0;
  mspace mem(x.e);
  int sites=data->length;
  if(st!=-1){
    sites=st;
  };
  long double *ans=new long double[x.l];
  for(int j=0;j<x.l;j++){
    *(ans+j)=0;
  };
  long double *siteliks=new long double[x.l];
//Store values currently being calculated here, to avoid corrupting
//the state in case of interruption
  i=getnext();
  int stv1=CurrentState->addvar('i',1,&i);
  CurrentState->addvar('L',x.l,ans,"hessian");//And to protect the state file.
  void **dbn=new void*[3];
  *dbn=this;
  *(dbn+1)=&ans;
  *(dbn+2)=&i;
  for(;i<sites;){//for each site.
    if(!((i+1)%showeverysite)){//should make this depend on the number of sites
      comm<<"Working on site "<<i+1<<" of "<<sites<<":\n";
      comm<<"In thread \""<<pthread_self()<<"\"\n";
    };
    this->sitehess(f,data,p,i,siteliks,x,mem);
    sigset_t all;
    sigset_t old;
    sigfillset(&all);
    sigprocmask(SIG_BLOCK,&all,&old);//Block signals while doing this
    accumulate(siteliks,ans,x.e,x.l,dbn);
    i=getnext();
    sigprocmask(SIG_SETMASK,&old,NULL);//Unblock signals again
  };
  delete[] dbn;
  CurrentState->remvar(stv1,"endhessian",1);
  delete[] siteliks;
  return ans;
};

void tree::sitehess(const Factmatrix& f,const sequence *data,const params& p,int i,long double *siteliks,const precalc &x,mspace &mem){
  //Calculates the hessian for the loglikelihood on the chosen site, i. 
  this->liklistdown(f,data,i,x.ex);
  this->liklistup(p,f,x.ex,NULL);
  //Start with likelihood
  *(siteliks)=0;
  for(int j=0;j<NumCodons;j++){
    *(siteliks)+=*(p.pi+j)*(*(down+j));
  };
  for(int j=1+x.e;j<x.l;j++){
    *(siteliks+j)=0;
  };
  this->liklistextra(f,p,x.ex,x.coeffs,x.dD,mem.T,mem.U);

    //Actually, the following few lines should be able to be run once for all parameters, saving time.
  long double *upvals=new long double[x.e*NumCodons]; 
  long double *downvals=new long double[x.e*NumCodons]; 
  /*    const long double **tmp=this->getuplists();
	f.gamma.MultiTranspActProd(e,tmp,p.pi,upvals);//Actually, these can be replaced.
	delete[] tmp;
	tmp=this->getdownlists();
    f.gammainv.MultiAct(e,tmp,downvals);
    delete[] tmp;  */
  this->getallup2lists(upvals);
  this->getalldown3lists(downvals);
  for(int j=0;j<NumCodons;j++){
    for(int k=0;k<NumCodons;k++){
      *(mem.coeffsums+j*NumCodons+k)=0;
      *(mem.coeffsums2+j*NumCodons+k)=0;
      *(mem.coeffsums3+j*NumCodons+k)=0;
      *(mem.coeffsums4+j*NumCodons+k)=0;
      *(mem.coeffsums5+j*NumCodons+k)=0;
    };
    *(mem.lsq+j)=0;
  };
  for(int ed=0;ed<x.e;ed++){	
    for(int j=0;j<NumCodons;j++){
      for(int k=0;k<NumCodons;k++){//These loops take up about 5% total time
	long double w=(*(upvals+ed*NumCodons+j))*(*(downvals+ed*NumCodons+k));
	*(mem.coeffsums+j*NumCodons+k)+=*(x.coeffs+ed*NumCodonsSq+j*NumCodons+k)*w;
	long double t=(*(x.ex+ed*NumCodons+j))*w;
	*(mem.coeffsums2+j*NumCodons+k)+=t;
	*(mem.coeffsums3+j*NumCodons+k)+=t*(*(x.le+ed));
	long double u=(*(x.ex+ed*NumCodons+k))*w;
	*(mem.coeffsums4+j*NumCodons+k)+=u;
	*(mem.coeffsums5+j*NumCodons+k)+=u*(*(x.le+ed));
	if(*(x.part+j)==*(x.part+k)){
	  *(mem.lsq+j)+=*(x.le+ed)*(*(x.le+ed))*(*(x.ex+ed*NumCodons+j))*w;
	};
      };
    };
  };
  //Up to here, this can all be precalculated.
  long double *jj=siteliks+1+x.e+params::numpars;
  for(int j=0;j<x.e;j++){
    //Second derivative with respect to branch j
    long double len=*(x.le+j);
    const tree *curr=this->findedge(j,len);
    *(jj++)=d2liketact(f,curr->up2,curr->down3,len,p,x.ex+j*NumCodons);
    this->secderbl(j,p,f,x.ex,jj,x.e,0);
    for(int k=0;k<params::numpars;k++){
      //Derivative with respect to branch j and parameter k
      *(jj++)=this->secderblp(j,k,p,f,x.ex,*(x.dD+k),0,upvals,x.coeffs);
    };
  };
  long double *acc=siteliks+1+x.e+params::numpars+x.e*params::numpars+x.e*(x.e+1)/2;
  int ii=0;
  for(int j=0;j<params::numpars;j++){
    for(int k=j;k<params::numpars;k++){
      const long double *ext=x.ex;
      const long double *cfs=x.coeffs;
      *(acc+ii)=this->secder(j,k,p,f,0,*(x.dD+j),*(x.dD+k),ext,cfs);
      *(acc+ii)+=this->sdsame(j,k,f,p,*(x.dD+j),*(x.dD+k),p.d2D(j,k),*(x.mats+4*ii),*(x.mats+4*ii+1),*(x.mats+4*ii+2),*(x.mats+4*ii+3),x.part,upvals,downvals,x.ex,x.le,x.FIJ,x.coeffs,mem.coeffsums2,mem.coeffsums3,mem.coeffsums4,mem.coeffsums5,mem.lsq,mem.T,mem.U);
      long double *ent=p.d2D(j,k).entries;
      for(int jj=0;jj<NumCodonsSq;jj++){
	*(acc+ii)+=*(ent+jj)*(*(mem.coeffsums+jj));
      };
      ii++;
    };
  };//Need to take derivatives with respect to branch lengths.
  for(int j=0;j<x.e;j++){//for each edge
    long double len;
    this->findedge(j,len);
    *(siteliks+j+1)=dliketact2(f,upvals+j*NumCodons,downvals+j*NumCodons,len,p,x.ex+j*NumCodons);
  };
  delete[] upvals;
  delete[] downvals;
  for(int k=0;k<params::numpars;k++){
      //Calculate first derivatives with respect to parameters.
    long double *lks=new long double[NumCodons];
    for(int kk=0;kk<NumCodons;kk++){
      *(lks+kk)=0;
    };
    for(int kk=0;kk<NumCodons;kk++){
      for(int ii=0;ii<n;ii++){
	long double th=1;	  
	for(int jj=0;jj<n;jj++){
	  if(jj==ii){	      
	    th*=*((subtree+ii)->extra+k*NumCodons+kk);
	  }else{
	    th*=*((subtree+jj)->down2+kk);
	  };
	};
	*(lks+kk)+=th;
      };
    };
    for(int kk=0;kk<NumCodons;kk++){
      *(siteliks+1+x.e+k)+=*(lks+kk)*(*(p.pi+kk));
    };
    delete[] lks;
  };
};


long double *tree::derivs(const Factmatrix& f,const sequence *data,const params& p){
  int e=this->edges();
  int l=e+params::numpars+1;
  int i=0;
  int sites=data->length;
  long double *siteliks=new long double[l*sites];
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  const Realmatrix *D=p.Derivmats();
  for(i=0;i<sites;i++){//for each site.
    this->liklistdown(f,data,i,ex);
    this->liklistup(p,f,ex,NULL);
    //Start with likelihood
    *(siteliks+i*l)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks+i*l)+=*(p.pi+j)*(*(down+j));
    };
    for(int j=1+e;j<l;j++){
      *(siteliks+i*l+j)=0;
    };
    for(int j=0;j<e;j++){//for each edge
      long double len;
      const tree* curr=this->findedge(j,len);
      *(siteliks+i*l+j+1)=dliketact(f,curr->up2,curr->down3,len,p,ex+j*NumCodons);
    };
    for(int k=0;k<params::numpars;k++){
      //Calculate first derivatives with respect to parameters.
      *(siteliks+i*l+k+e+1)=0;
      for(int j=0;j<e;j++){
	long double len;
	const tree* curr=this->findedge(j,len);
	*(siteliks+i*l+e+k+1)+=dlikeqact(f,curr->up2,curr->down3,len,p,k,*(D+k),ex+j*NumCodons);
      };
    };
  };
  delete[] ex;
  delete[] le;

  long double *ans=new long double[l];
  *ans=0;
  for(int j=0;j<sites;j++){
    *ans+=log(*(siteliks+j*l));
  };
  for(int i=1;i<1+e+params::numpars;i++){
    *(ans+i)=0;
    for(int s=0;s<sites;s++){
      *(ans+i)+=(*(siteliks+s*l+i))/(*(siteliks+s*l));
    };
  };
  delete[] siteliks;
  return ans;
};


long double *tree::approxHessian(const Factmatrix& f,const sequence *data,const params& p){
  int e=this->edges();
  int l=e+params::numpars+1;
  int i=0;
  int sites=data->length;
  long double *siteliks=new long double[l*sites];
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  const Realmatrix *D=p.Derivmats();
  for(i=0;i<sites;i++){//for each site.
    this->liklistdown(f,data,i,ex);
    this->liklistup(p,f,ex,NULL);
    //Start with likelihood
    *(siteliks+i*l)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks+i*l)+=*(p.pi+j)*(*(down+j));
    };
    for(int j=1+e;j<l;j++){
      *(siteliks+i*l+j)=0;
    };
    for(int j=0;j<e;j++){//for each edge
      long double len;
      const tree* curr=this->findedge(j,len);
      *(siteliks+i*l+j+1)=dliketact(f,curr->up2,curr->down3,len,p,ex+j*NumCodons);
    };
    for(int k=0;k<params::numpars;k++){
      //Calculate first derivatives with respect to parameters.
      *(siteliks+i*l+k+e+1)=0;
      for(int j=0;j<e;j++){
	long double len;
	const tree* curr=this->findedge(j,len);
	*(siteliks+i*l+e+k+1)+=dlikeqact(f,curr->up2,curr->down3,len,p,k,*(D+k),ex+j*NumCodons);
      };
    };
  };
  delete[] ex;
  delete[] le;

  long double *ans=new long double[l*(l+1)/2];
  *ans=0;
  for(int j=0;j<sites;j++){
    *ans+=log(*(siteliks+j*l));
  };
  for(int i=1;i<1+e+params::numpars;i++){
    *(ans+i)=0;
    for(int s=0;s<sites;s++){
      *(ans+i)+=(*(siteliks+s*l+i))/(*(siteliks+s*l));
    };
  };
  int v=1+e+params::numpars;
  for(int j=0;j<e+params::numpars;j++){
    for(int k=j;k<e+params::numpars;k++){
      *(ans+v)=*(ans+j+1)*(*(ans+k+1));
      for(int s=0;s<sites;s++){
	*(ans+v)-=(*(siteliks+s*l+j+1))*(*(siteliks+s*l+k+1));
      };
      v++;
    };
  };
  delete[] siteliks;
  return ans;
};

long double tree::likelihood(const Factmatrix& f,const sequence *data,const params& p){
  int i=0;
  int sites=data->length;
  int e=(this->edges());
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  long double *siteliks=new long double[sites];
  for(i=0;i<sites;i++){//for each site.
    this->liklistdown(f,data,i,ex);
    *(siteliks+i)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks+i)+=*(p.pi+j)*(*(down+j));
    };
    if(*(siteliks+i)<=0){
      this->debugneg(*(siteliks+i),cout,i);
    };
  };
  long double ans=0;
  for(int j=0;j<sites;j++){
    ans+=log(*(siteliks+j));
  };
  delete[] ex;
  delete[] le;
  delete[] siteliks;
  return ans;
};

long double tree::fastlikelihood(const Factmatrix& f,const sequence *data,const params& p){
  int i=0;
  int sites=data->length;
  int e=(this->edges());
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  Realmatrix gis(NumCodons);
  for(int i=0;i<NumCodonsSq;i++){
    *(gis.entries+i)=*(f.gamma.entries+i)*(*(f.gamma.entries+i));
  };
  long double *siteliks=new long double[sites];
  for(i=0;i<sites;i++){//for each site.
    this->liklistdownfast(f,data,i,ex,gis);
    *(siteliks+i)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks+i)+=*(p.pi+j)*(*(down+j));
    };
  };
  long double ans=0;
  for(int j=0;j<sites;j++){
    ans+=log(*(siteliks+j));
  };
  delete[] ex;
  delete[] le;
  delete[] siteliks;
  return ans;
};

long double *tree::sitelikelihoods(const Factmatrix& f,const sequence *data,const params& p){
  //Outputs the log likelihood of each site.
  int i=0;
  int sites=data->length;
  long double *ans=new long double[sites];
  int e=(this->edges());
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  for(i=0;i<sites;i++){//for each site.
    this->liklistdown(f,data,i,ex);
    *(ans+i)=0;
    for(int j=0;j<NumCodons;j++){
      *(ans+i)+=*(p.pi+j)*(*(down+j));
    };
    if(*(ans+i)<=0){
      this->debugneg(*(ans+i),cout,i);
    };
    *(ans+i)=log(*(ans+i));
  };
  delete[] ex;
  delete[] le;
  return ans;
};


long double *tree::directionalDerivs(const Factmatrix& f,const sequence *data,const params& p,const long double *dir){
  //Calculates the first and second directional derivatives in the
  //direction dir. 

  //Need to think about this.
  int e=this->edges();
  int i=0;
  int sites=data->length;
  long double *siteliks=new long double[3*sites];
  long double *ex=new long double[NumCodons*e];
  long double *le=this->getlengths();
  long double *coeffs=NULL;
  for(int i=0;i<e;i++){
    for(int j=0;j<NumCodons;j++){
      *(ex+i*NumCodons+j)=exp((*(le+i))*(*(f.D.entries+j)));//Should only calculate this matrix once.
    };
  };
  Realmatrix X(NumCodons);
  Realmatrix X2(NumCodons);
  for(int i=0;i<NumCodonsSq;i++){
    *(X.entries+i)=0;
    *(X2.entries+i)=0;
  };
  const Realmatrix *dD=p.Derivmats();
  for(int i=0;i<params::numpars;i++){
    for(int j=0;j<NumCodonsSq;j++){
      *(X.entries+j)+=*(dir+e+i)*(*((dD+i)->entries+j));
    };
  };
  for(int i=0;i<params::numpars;i++){
    for(int j=0;j<params::numpars;j++){
      const Realmatrix& O=p.d2D(i,j);
      for(int k=0;k<NumCodonsSq;k++){
	*(X2.entries+k)+=*(dir+e+i)*(*(dir+e+j))*(*(O.entries+k));
      };
    };
  };
  int *part=partition(NumCodons,f.D.entries);
  long double **mats=new long double*[4];
  *(mats)=new long double[NumCodonsSq];    
  *(mats+1)=new long double[NumCodonsSq];    
  *(mats+2)=new long double[NumCodonsSq];    
  *(mats+3)=new long double[NumCodonsSq];    
  makematricesdir(X,f.D.entries,part,*(mats),*(mats+1),*(mats+2),*(mats+3));
  long double *FIJ=new long double[NumCodonsSq];
  for(int i=0;i<NumCodons;i++){
    for(int j=0;j<NumCodons;j++){
      *(FIJ+i*NumCodons+j)=(*(part+i)==*(part+j))?0:1/(*(f.D.entries+i)-*(f.D.entries+j));
    };
  };
  for(i=0;i<sites;i++){//for each site.
    this->liklistdown(f,data,i,ex);
    this->liklistup(p,f,ex,NULL);
    //Start with likelihood
    *(siteliks+i*3)=0;
    *(siteliks+i*3+1)=0;
    *(siteliks+i*3+2)=0;
    for(int j=0;j<NumCodons;j++){
      *(siteliks+i*3)+=*(p.pi+j)*(*(down+j));
    };
    for(int j=0;j<e;j++){//for each edge
      long double len;
      const tree* curr=this->findedge(j,len);
      *(siteliks+i*3+1)+=*(dir+e)*dliketact(f,curr->up2,curr->down3,len,p,ex+j*NumCodons);
    };
    for(int j=0;j<e;j++){
      long double len;
      const tree* curr=this->findedge(j,len);
      *(siteliks+i*3+1)+=dlikeqact2(f,curr->up2,curr->down3,len,p,X,ex+j*NumCodons);
    };
    this->liklistextradir(f,p,ex,X,dir);
    //Actually, the following few lines should be able to be run once for all parameters, saving time.
    long double *upvals=new long double[e*NumCodons]; 
    long double *downvals=new long double[e*NumCodons]; 
    //    const long double **tmp=this->getuplists();
    //    f.gamma.MultiTranspActProd(e,tmp,p.pi,upvals);//Actually, these can be replaced.
    //    delete[] tmp;
    //    tmp=this->getdownlists();
    //    f.gammainv.MultiAct(e,tmp,downvals);
    //    delete[] tmp;  
    this->getallup2lists(upvals);
    this->getalldown3lists(downvals);
  //Up to here, this can all be precalculated.
    const long double *ext=ex;
    *(siteliks+i*3+2)+=this->secderbldir(dir,p,f,ex,0);
    *(siteliks+i*3+2)+=this->secderblpdir(dir,p,f,ex,X,0,coeffs);
    *(siteliks+i*3+2)+=this->secderdir(p,f,0,X,ext,coeffs);
    *(siteliks+i*3+2)+=this->sdsamedir(f,p,X,X2,*(mats),*(mats+1),*(mats+2),*(mats+3),part,upvals,downvals,ex,le,FIJ);
    delete[] upvals;
    delete[] downvals;
  };
  delete[] ex;
  delete[] le;
  delete[] *mats;
  delete[] *(mats+1);
  delete[] *(mats+2);
  delete[] mats;
  delete[] part;
  delete[] FIJ;
  long double *ans=new long double[3];
  *ans=0;
  *(ans+1)=0;
  for(int j=0;j<sites;j++){
    *ans+=log(*(siteliks+j*3));
  };
  for(int s=0;s<sites;s++){
    *(ans+1)+=(*(siteliks+s*3+1))/(*(siteliks+s*3));
  };
  *(ans+2)=0;//*(ans+1)*(*(ans+1));
  for(int s=0;s<sites;s++){
    *(ans+2)+=(*(siteliks+s*3+2))/(*(siteliks+s*3));
    *(ans+2)-=(*(siteliks+s*3+1))*(*(siteliks+s*3+1))/(*(siteliks+s*3)*(*(siteliks+s*3)));
  };
  delete[] siteliks;
  return ans;
};

void dneg(void *d){
  tree *t=(tree *)*(void **)d;
  long double val=*(long double *)(*((void **)d+1));
  int no=*(int *)(*((void **)d+2));
  t->debugneg(val,cout,no);
};

#include "Mixture.cpp"
